from Providers import Provider
from ComputedAttribute import ComputedAttribute
from DataSkins import AddedStatus, ChangedStatus, DeletedStatus, NOT_FOUND
# XXX from zLOG import LOG, INFO
from PlugIns import defaultConstructors
from Expressions import *
from string import split,strip

def _id(client, id=id): return id(getattr(client,'aq_base',client))
































class Agent(Provider):

    """Thing that monitors DataSkin events"""
    
    __plugin_kind__ = 'Monitoring Agent'


    def _v_tranlog(self):
        l = self._v_tranlog = {}
        return l


    _v_tranlog = ComputedAttribute(_v_tranlog)


    def namesForRegistration(self,container):
        "Agents provide 'handlers'; subclasses may add attributes and sheets"
        return { 'provides': ('handlers',) }


    def _objectAdding(self,client, _id=_id):

        i = _id(client)
        l = self._v_tranlog
        c = l.get(i)

        if c is None:
            l[i] = AddedStatus, client, None
            return

        s,c,m = c

        if s is DeletedStatus:
            l[i] = ChangedStatus, client, m

        return





    def _objectDeleting(self,client, _id=_id):

        i = _id(client)
        l = self._v_tranlog
        c = l.get(i)

        if c is None:
            l[i] = DeletedStatus, client, self._getMementoFor(client)
            return

        s,c,m = c

        if s is AddedStatus:
            del l[i]

        elif s is ChangedStatus:
            l[i] = DeletedStatus, client, m


    def _objectChanging(self,client, _id=_id):

        i = _id(client)
        l = self._v_tranlog
        c = l.get(i)

        if c is None:
            l[i] = ChangedStatus, client, self._getMementoFor(client)














    def _checkpoint(self):

        # Send 'final' versions of events

        try:
        
            for s,c,m in self._v_tranlog.values():
            
                if s is ChangedStatus:
                    self._objectChanged(c,m)
                    
                elif s is AddedStatus:
                    self._objectAdded(c)
                    
                else:
                    self._objectDeleted(c,m)
                    
        finally:
            self._v_tranlog.clear()


    def _revert(self):
        self._v_tranlog.clear()


    def _cleanup(self):
        self._v_tranlog.clear()














    def _getMementoFor(self,client):
        # Override to return a memento object relevant to the agent
        pass

    def _objectChanged(self,client,memento):
        pass

    def _objectDeleted(self,client,memento):
        pass

    def _objectAdded(self,client):
        pass





























class GenericTrigger(Agent):
    """Call an expression upon change to an object"""
    
    meta_type = 'Generic Trigger'
    
    _mementos = memento_exprs = handle_set_for = ()
    _expr = None
    callexpr = ''
    applicable_events = event_kinds = 'ADD', 'CHANGE', 'DELETE'
    
    def _propertiesChanged(self):
        m = self._mementos = []
        
        if self.callexpr:
            self._expr = Expression(self.callexpr)
        else:
            self._expr = None

        for l in self.memento_exprs:
            if strip(l):
                if '=' in l:
                    a,e = split(strip(l), '=', 1)
                    m.append(a,Expression(e))
                else:
                    m.append(l,Name(l))

    _properties=(
        {'id':'title', 'type': 'string', 'mode': 'w'},
        {'id':'applicable_events', 'type': 'multiple selection',
            'mode': 'w', 'select_variable':'event_kinds'},
        {'id':'callexpr', 'type': 'string', 'mode': 'w'},
        {'id':'memento_exprs', 'type': 'lines', 'mode': 'w'},
        {'id':'handle_set_for','type': 'tokens', 'mode': 'w'},
    )
    propertyLabels = {
        'title':'Title',
        'applicable_events': 'Upon:',
        'callexpr':'Execute:',
        'memento_exprs': 'Keeping:', 'handle_set_for':'Set&nbsp;Attrs:'
    }

    def _getMementoFor(self,client):
    
        if not self._mementos:
            return _emptyMemento
            
        memento = _memento()
        d=memento.__dict__

        t = NamespaceStack(self.REQUEST)
        t._push(InstanceDict(client, t))
        
        try:
            for n,e in self._mementos:
                d[n] = e.eval(t)                
        finally:
            t._pop()

        return memento

    def _objectChanged(self,client,memento):
        self._notify('CHANGE',client,memento)

    def _objectDeleted(self,client,memento):
        self._notify('DELETE',client,memento)

    def _objectAdded(self,client):
        self._notify('ADD',client,_emptyMemento)

    def namesForRegistration(self,container):
        if self.handle_set_for:
            return {
                'provides': ('handlers','attributes'),
                'setattr': self.handle_set_for,
                'delattr': self.handle_set_for,
            }
    
        return { 'provides': ('handlers',) }
        



    def _notify(self,event,client,memento):

        if event not in self.applicable_events or self._expr is None:
            return
            
        #LOG('ZPatterns', INFO, 'Object %s from [%s] to [%s]' %
        #    (event,memento.__dict__,self._getMementoFor(client).__dict__)
        #) XXX 

        t = NamespaceStack(self.REQUEST)
        changed = client._v_changedAttrs_

        t._push(InstanceDict(self, t))
        t._push({
            'self':client, 'TRIGGER_EVENT':event, 'OLD':memento,
            'CHANGED': changed.has_key, 'CHANGED_ATTRS': changed.keys,
        })
                
        try:
            result = self._expr.eval(t)
        finally:
            t._pop(2)

        return result


    def _SetAttributeFor(self,client,name,value):
        client._getCache()[name]=value
        return 1

    def _DelAttributeFor(self,client,name):
        """Delete the attribute and return true if successful"""
        client._getCache()[name]=NOT_FOUND
        return 1


class _memento: pass
_emptyMemento=_memento()

ExpressionTrigger=GenericTrigger

def initialize(context):

    context.registerPlugInClass(
        GenericTrigger,
        permission = 'Add Generic Triggers',
        constructors = defaultConstructors(GenericTrigger,globals()),
        icon = 'www/trigger.gif'
    )

