from PlugIns import defaultConstructors
from Providers import Provider
from Globals import default__class_init__
from string import split,strip
from Expressions import *
import sys
from zLOG import LOG, WARNING


































class AttributeProvider(Provider):
    """
    AttributeProvider example class.  This attribute provider assumes that
    the object containing the attributes is persistent and can save
    whatever is desired.  All of its methods would need to be redefined
    for providers that implement fixed-schema and/or non-persistence-based
    attributes.
    """
    __plugin_kind__ = 'Attribute Provider'

    def namesForRegistration(self,container):
        """
Return a tuple (readnames,writenames) where each is a sequence of attribute
names which this attribute provider supports reading and writing, respectively.
'*' may be included in either or both sequences, which means the provider will
be called if a more specific provider could not be found for a given attribute.
This is so that attribute providers can potentially create attributes
dynamically based on the requested name.
        """
        return {
            'provides':('attributes',),
            'setattr': self.Attributes, 'delattr': self.Attributes
        }

    def _AttributeFor(self,client,name,default=None):
        """
        Since persistent attributes are stored in the object, if we get
        here, then we are trying to get something that doesn't exist!
        """
        return default

    def _SetAttributeFor(self,client,name,value):
        """Set the attribute and return true if successful"""
        client.__dict__[name]=value; client._p_changed = 1
        return 1

    def _DelAttributeFor(self,client,name):
        """Delete the attribute and return true if successful"""
        del client.__dict__[name]; client._p_changed = 1
        return 1

    # Class Metadata
    
    meta_type='Persistent Internal Attribute Provider'
    
    Attributes = ('*',)

    _properties=(
        {'id':'title', 'type': 'string', 'mode': 'w'},
        {'id':'Attributes', 'type': 'tokens', 'mode': 'w'},
    )































class AcquiredAttributeProvider(AttributeProvider):

    """Provide attributes from parent ProviderContainer"""

    def namesForRegistration(self,container):
        c=container.aq_parent
        r=getattr(c,'_getRegistry',None)

        if r is None:
            raise TypeError, "No providers available to acquire in /%s" % c.absolute_url(1)

        return {
            'provides': ('attributes',),
            
            'getattr':
                self._handledAttrs( r('getattr').keys() ),
                
            'setattr':
                self._handledAttrs( r('setattr').keys() ),

            'delattr':
                self._handledAttrs( r('delattr').keys() )
        }
        
    def _handledAttrs(self, names):
        a = self.Attributes
        if '*' in a:     return names
        if '*' in names: return a
        return filter(lambda x,a=a: x in a, names)

    def _AttributeFor(self,client,name,default=None):
        return self.aq_parent.aq_parent._AttributeFor(client,name,default)

    def _SetAttributeFor(self,client,name,value):
        return self.aq_parent.aq_parent._SetAttributeFor(client,name,value)

    def _DelAttributeFor(self,client,name):
        return self.aq_parent.aq_parent._DelAttributeFor(client,name)



    def _objectAdding(self,client):
        return self.aq_parent.aq_parent._objectAdding(client)

    def _objectDeleting(self,client):
        return self.aq_parent.aq_parent._objectDeleting(client)

    meta_type='Acquired Attribute Provider'


































class GenericAttributeProvider(AttributeProvider):
    """
    XXX Explain this
    """

    Attributes = ()
    
    def _propertiesChanged(self):
    
        self.Attributes = []

        if self.fromexpr:
            self._fromex = Expression(self.fromexpr)
        else:
            self._fromex = None
            
        for l in filter(None,map(strip,self.attrsexprs)):
            if '=' in l:
                a,e = split(l,'=',1)
                self.Attributes.append(a,Expression(e))
            else:
                self.Attributes.append(l,Name(l,0))


    def namesForRegistration(self,container):
        """XXX"""
            
        return {
            'provides':('attributes',),
            'getattr': map(lambda x: x[0], self.Attributes)
        }










    def _AttributeFor(self,client,name,default=None):
    
        try:
            t = NamespaceStack(self.REQUEST)
            
            data = {'self':client, 'ATTRIBUTE_NAME':name, 'NOT_FOUND':default}
	
            if self.fromexpr:

                t._push(InstanceDict(self, t))
                t._push(data)
                
                try:
                    result = data['RESULT'] = self._fromex.eval(t)
                finally:
                    t._pop(2)

            else:
                result = data['RESULT'] = client

            if result is default:
                return default
            
            t._push(InstanceDict(result, t))
            t._push(data)
        
            try:
                c = client._getCache()
                for a,e in self.Attributes:
                    c[a] = e.eval(t)
            finally:
                    t._pop(2)
            
            return c.get(name,default)

        except:
            LOG('ZPatterns',WARNING,('Error computing attribute %s' % name),
                '', sys.exc_info(), 1
            )


    # Class Metadata
    
    meta_type='Generic Attribute Provider'
    
    fromexpr=''
    attrsexprs=[]

    _properties=(
        {'id':'title', 'type': 'string', 'mode': 'w'},
        {'id':'fromexpr', 'type': 'string', 'mode': 'w'},
        {'id':'attrsexprs', 'type': 'lines', 'mode': 'w'},
    )

ExprGetterAttributeProvider = GenericAttributeProvider



























def initialize(context):

    context.registerPlugInClass(
        GenericAttributeProvider,
        permission = 'Add Generic Attribute Providers',
        constructors = defaultConstructors(GenericAttributeProvider,
                        globals()),
        icon = 'www/exprgetter.gif'
    )

    context.registerPlugInClass(
        AttributeProvider,
        permission = 'Add Attribute Providers',
        constructors = defaultConstructors(AttributeProvider,globals()),
        icon = 'www/attrprov.gif'
    )

    #context.registerPlugInClass(
    #    AcquiredAttributeProvider,
    #    permission = 'Add Acquired Attribute Providers',
    #    constructors = defaultConstructors(AcquiredAttributeProvider,globals()),
    #    icon = 'www/acqattrprov.gif'
    #)



