from Providers import ProviderGroup
from PlugIns import PlugInContainer
from Globals import HTMLFile, default__class_init__, PersistentMapping
from OFS.Folder import Folder
from Transactions import Transactional
from ComputedAttribute import ComputedAttribute
from AttributeProviders import AttributeProvider
from SheetProviders import SheetProvider
from ExtensionClass import Base
import Products

try:
    from thread import get_ident
except:
    def get_ident(): pass

_marker = []

class DataManager(Transactional, PlugInContainer):

    """Something which provides objects with data"""

    def _objectChanging(self,client):
        for ob in self._getProviders('handlers'):
            ob.__of__(self)._objectChanging(client)

    def _objectAdding(self,client):
        for ob in self._uniqueProviders(('attributes','sheets','handlers')):
            ob.__of__(self)._objectAdding(client)

    def _objectDeleting(self,client):
        for ob in self._uniqueProviders(('handlers','sheets','attributes')):
            ob.__of__(self)._objectDeleting(client)








    def _v_canonicals(self):
        l = self._v_canonicals = {}
        return l

    _v_canonicals = ComputedAttribute(_v_canonicals)

    def _canonicalForm(self,key,default=None):
        return self._v_canonicals.get(key,default)

    def _registerCanonical(self, key, canonical):
        self._register()
        self._v_canonicals[key] = canonical
    
    def _getDataManagerFor(self,client,default=None):
        return self.aq_inner

    def _readableSlotFor(self,client):
        slot = getattr(client,'__skinSlot__',_marker)

        if slot is _marker: return {}

        client._setSlot(slot); return slot

    def _writeableSlotFor(self,client):
        slot = getattr(client,'__skinSlot__',_marker)

        if slot is _marker:
            slot = client.__skinSlot__ = PersistentMapping()

        client._setSlot(slot); return slot











    def _checkpoint(self):
        for ob in self.__dataProviders__:
            ob.__of__(self)._checkpoint()

    def _vote(self):
        for ob in self.__dataProviders__:
            ob.__of__(self)._vote()

    def _finalize(self):
        for ob in self.__dataProviders__:
            ob.__of__(self)._finalize()


    def _revert(self):
        for ob in self.__dataProviders__:
            try:    ob.__of__(self)._revert()
            except: pass

    def _rollback(self):
        for ob in self.__dataProviders__:
            try:    ob.__of__(self)._rollback()
            except: pass


    def _cleanup(self, type = type):
        for ob in self.__dataProviders__:
            try:    ob.__of__(self)._cleanup()
            except: pass

        tid = type(id(self))
        
        for key, client in self._v_canonicals.items():
            if type(key) is tid: 
                client._clearPerTransactionCache()
                
        self._v_canonicals.clear()





    __provider_registry__ = {}
    __dataProviders__ = ()
    
    DataPlugins = ProviderGroup(
        'DataPlugins',
        ['Property Sheet Provider','Attribute Provider', 'Monitoring Agent',
          'Data Plug-in'],
        title = 'Data Plug-ins',
        attr  = '__dataProviders__'
    )

    __plugin_groups__ = (DataPlugins,)
    
    def _getRegistry(self,name):
        return self.__provider_registry__.get(name,{})

    def _getProviders(self,what):
        return self._getRegistry('provides').get(what,())

    def _uniqueProviders(self,kinds):
        d = {}; have = d.has_key
        l = []
        providers = self._getRegistry('provides')
        
        for kind in kinds:
            for pp in providers.get(kind,()):
                if have(id(pp)): continue
                l.append(pp); d[id(pp)]=1
                
        return l











    def _zclassOK(self, z):
        """Verify that the ZClass selected is acceptable for this DataManager"""
        # All we check here is rack-mountability/dataskinnishness
        return getattr(z._zclass_,'_isRackMountable',0)


    def _unifiedZClassRegistry(self):
        """Return a sanely unified registry of ZClasses and ZClass bases"""

        z = Products.meta_classes
        i = Products.meta_class_info

        r = {}
        for k in z.keys():
            r[k] = i[k], z[k]

        for data in self.aq_acquire('_getProductRegistryData')('zclasses'):
            path = '%(product)s/%(id)s' % data
            r[path] = '%(product)s: %(id)s' % data, data['meta_class']

        for k,(i,z) in r.items():
            if not self._zclassOK(z): del r[k]

        return r

    def manage_product_zclass_info(self):
        """Return a list of ZClasses suitable for use in this DataManager"""

        return map(lambda x, sel=self._isSelected: {
                'name':x[1][0],
                'path':x[0],
                'meta_type': x[1][1]._zclass_.meta_type,
                'selected': sel(x[1][0],x[0],x[1][1]) and 'selected' or '',
            }, 
            self._unifiedZClassRegistry().items()
        )

    def _isSelected(self,name,path,klass): pass



    def __setstate__(self,state):
        """Upgrade data structures from pre 0.4.0 versions"""
        
        d = self.__dict__; d.clear()
        
        if state.has_key('__provider_registry__'):
            d.update(state)
            return

        p = {}
        l = state['__dataProviders__'] = []
        r = state['__provider_registry__'] = {'provides':p}

        for new,old in (
                ('sheets',    '__sheetProviders__'),
                ('attributes','__attributeProviders__'),
                ('handlers','__indexAgents__'),
                ('handlers','__triggerAgents__'),
            ):
            if state.has_key(old):
                p[new] = state[old]; l.extend(state[old])
                del state[old]

        for new,old in (
                ('getattr', '__getattr_registry__'),
                ('setattr', '__setattr_registry__'),
                ('sheet_names', '__sheet_registry__'),
                ('sheet_xmlns', '__xmlns_registry__'),
            ):
            if state.has_key(old):
                r[new] = state[old]
                del state[old]

        for ap in 'getattr','setattr':
            if r.has_key(ap):
                z = r[ap]
                for k,v in z.items(): z[k] = [v]    # new registry expects provider list
                
        if r.has_key('setattr'): r['delattr'] = r['setattr'].copy()       
        d.update(state)        
           
    def _AttributeFor(self, client, name, default):
        d = self._getRegistry('getattr')
        pp = d.get(name,_marker)
        if pp is _marker: pp=d.get('*',_marker)
 
        if pp is not _marker:
            for ap in pp:
                ap = ap.__of__(self)
                v = ap._AttributeFor(client,name,default)
                if v is not default: return v
            
        return default


    def _SetAttributeFor(self, client, name, value):
        d = self._getRegistry('setattr')
        pp = d.get(name,_marker)
        if pp is _marker: pp=d.get('*',_marker)

        if pp is not _marker:
            for ap in pp:
                ap = ap.__of__(self)
                v = ap._SetAttributeFor(client,name,value)
                if v: return v


    def _DelAttributeFor(self, client, name):
        d = self._getRegistry('delattr')
        pp = d.get(name,_marker)
        if pp is _marker: pp=d.get('*',_marker)

        if pp is not _marker:
            for ap in pp:
                ap = ap.__of__(self)
                v = ap._DelAttributeFor(client,name)
                if v: return v





    def _PropertySheetFor(self, client, name='', xmlns=''):
        return self._delegateProviders(client,name,xmlns,'_PropertySheetFor')


    def _addPropertySheetFor(self,client,name='',xmlns=''):
        """
Return client's new propertysheet matching name or xmlns, if possible.
Otherwise return None
        """
        return self._delegateProviders(client,name,xmlns,'_addPropertySheetFor')


    def _delPropertySheetFor(self,client,name='',xmlns=''):
        """
Delete client's propertysheet matching name or xmlns, if present.  Always
returns None (unless an unexpected error occurs, of course)
        """
        return self._delegateProviders(client,name,xmlns,'_delPropertySheetFor')


    def _delegateProviders(self,client,name,xmlns,mname):

        if xmlns:  k,d = xmlns, self._getRegistry('sheet_xmlns')
        elif name: k,d = name,  self._getRegistry('sheet_names')
        else:      return None

        for sp in d.get(k,[]) + d.get('*',[]):
            sp = sp.__of__(self)
            s = getattr(sp,mname)(client,name=name,xmlns=xmlns)
            if s is not None: return s











    def _PropertySheetsFor(self, client):
    
        """Return a (possibly empty) list of PropertySheet objects for client"""
        
        l = []
        
        for sp in self._getProviders('sheets'):
            sp = sp.__of__(self)
            l.extend(list(sp._PropertySheetsFor(client)))

        return l            


    __ac_permissions__ = (
        ('View management screens',
            ('manage_product_zclass_info',),
        ),
    )

default__class_init__(DataManager)





















_nullDMs = {}

class DefaultDataManager(Base):

    """Thing that returns a canonical NullDataManager for each thread"""
    
    def __of__(self,parent):
    
        id  = get_ident()
        NDM = _nullDMs.get(id)
        
        if NDM is None:
            NDM = _nullDMs[id] = NullDataManager(id='_DefaultDataManager')
            
        return NDM.__of__(parent)
    

class NullDataManager(DataManager):

    """DataManager that emulates ordinary Zope reality"""

    _AttributeFor    = AttributeProvider._AttributeFor.im_func
    _SetAttributeFor = AttributeProvider._SetAttributeFor.im_func
    _DelAttributeFor = AttributeProvider._DelAttributeFor.im_func

    _PropertySheetFor    = SheetProvider._PropertySheetFor.im_func
    _addPropertySheetFor = SheetProvider._addPropertySheetFor.im_func
    _delPropertySheetFor = SheetProvider._delPropertySheetFor.im_func
    _PropertySheetsFor   = SheetProvider._PropertySheetsFor.im_func
    _objectDeleting      = SheetProvider._objectDeleting.im_func

    def _objectChanging(self,client): pass
    def _objectAdding(self,client):   pass
    
