from DynPersist import DynPersist
from Globals import Persistent
from ComputedAttribute import ComputedAttribute
from Acquisition import ImplicitAcquisitionWrapper
from DataManagers import DefaultDataManager
from PropertySheets import VirtualSheets
from Globals import default__class_init__
from OFS.ObjectManager import ObjectManager
from OFS.CopySupport import CopySource

_marker = []

NOT_FOUND = []

_v_dm_ = '_v_dm_'
_v_status_ = '_v_status_'

ChangedStatus = 'Changed'
DeletedStatus = 'Deleted'
AddedStatus   = 'Added'

manage_afterAdd     = ObjectManager.manage_afterAdd.im_func
manage_beforeDelete = ObjectManager.manage_beforeDelete.im_func
manage_afterClone   = ObjectManager.manage_afterClone.im_func

















class DataSkin(DynPersist,Persistent):

    """Mixin for objects that want to delegate storage issues externally"""

    _rackMounted = 0
    _isRackMountable = 1


    def __init__(self, id):
        # Bypass attribute providers for setting up initial id, 
        # since they need the id in order to work!
        self.__dict__['id']=id; self._p_changed = 1


    def _setRack(self, rack):
        """Called by the rack before returning the object"""
        d = self.__dict__
        d[_v_dm_] = rack; d['_rackMounted'] = 1


    def _setSlot(self, slot):
        """
        Set up the slot object for persistent storage of object
        'baggage' (e.g. WebDAV property sheets, etc.).  Called by
        ZODB-based racks at object retrieval, or by non-ZODB based
        racks at the object's request.
        """
        self._v_readableSlot = self._v_writeableSlot = slot


    def commitSubtransaction(self): return get_transaction.commit(1)










    # Computed attributes to handle on-demand slot retrieval/allocation

    __per_transaction_cache_attrs__ = (
        '_v_currentSheets', '_v_attrCache', '_v_status_', '_v_dm_',
        '_v_changedAttrs_'
    )

    def _clearPerTransactionCache(self):
        """Get rid of stale data from previous transactions"""
        d=self.__dict__; have=d.has_key
        for a in self.__per_transaction_cache_attrs__:
            if have(a): del d[a]

    def _getCache(self):
        """Return attribute cache"""
        return self._v_attrCache
        
    def _v_readableSlot(self,_v_dm_=_v_dm_):
        return self.__dict__[_v_dm_]._readableSlotFor(self)

    def _v_writeableSlot(self,_v_dm_=_v_dm_):
        return self.__dict__[_v_dm_]._writeableSlotFor(self)

    def _v_currentSheets(self,_v_dm_=_v_dm_):
        """PropertySheets tuple (analagous to __propsets__)"""
        self._v_currentSheets = tuple(
            self.__dict__[_v_dm_]._PropertySheetsFor(self)
        )
        return self._v_currentSheets

    def _v_attrCache(self):
        """Cache of external attribute values"""
        v = self._v_attrCache = {}
        return v

    _v_readableSlot = ComputedAttribute(_v_readableSlot)
    _v_writeableSlot = ComputedAttribute(_v_writeableSlot)
    _v_currentSheets = ComputedAttribute(_v_currentSheets)
    _v_attrCache = ComputedAttribute(_v_attrCache)


    # Event management (avoids sequential repeats of same events)

    def _v_changedAttrs_(self):
        """Cache of changed attribute names"""
        v = self._v_changedAttrs_ = {}
        return v
        
    _v_changedAttrs_ = ComputedAttribute(_v_changedAttrs_)
    
    def _objectChanging(self,name,_v_status_=_v_status_,_v_dm_=_v_dm_):
    
        d=self.__dict__
        
        if self._v_status_ is not ChangedStatus:
            d[_v_dm_]._objectChanging(self)
            d[_v_status_] = ChangedStatus
        
        v = self._v_attrCache
        if v.has_key(name):
            del v[name]

        if d.has_key(name):
            del d[name]; self._p_changed=1

        self._v_changedAttrs_[name]=1


    def _objectAdding(self,_v_status_=_v_status_,_v_dm_=_v_dm_):
        if self._v_status_ is not AddedStatus:
            d=self.__dict__
            d[_v_dm_]._objectAdding(self)
            d[_v_status_] = AddedStatus
        
        
    def _objectDeleting(self,_v_status_=_v_status_,_v_dm_=_v_dm_):
        if self._v_status_ is not DeletedStatus:
            d=self.__dict__
            d[_v_dm_]._objectDeleting(self)
            d[_v_status_] = DeletedStatus


    # Attribute handling

    def __get_attr__(self,name,_v_dm_=_v_dm_):

        cache = self._v_attrCache

        v = cache.get(name,_marker)

        if v is _marker:
            dm = self.__dict__[_v_dm_]
            self = dm._canonicalForm(id(self),self)

            v = dm._AttributeFor(self,name,NOT_FOUND)
            if v is NOT_FOUND:
                v = getattr(self.__class__,'class_default_for_'+name,NOT_FOUND)

            cache[name]=v
    
        if v is not NOT_FOUND:
            return v

        raise AttributeError, name


    def __set_attr__(self,name,val,_v_dm_=_v_dm_):
        dm = self.__dict__[_v_dm_]
        self = dm._canonicalForm(id(self),self)
        self._objectChanging(name)    
        
        if not dm._SetAttributeFor(self,name,val):
            raise AttributeError, name


    def __del_attr__(self,name,_v_dm_=_v_dm_):
        dm = self.__dict__[_v_dm_]
        self = dm._canonicalForm(id(self),self)
        self._objectChanging(name)

        if not dm._DelAttributeFor(self,name):
            raise AttributeError, name

    # DataManager hookup

    _DefaultDataManager = DefaultDataManager()
    
    def __of__(self,parent,
        _iaw=ImplicitAcquisitionWrapper,_v_dm_=_v_dm_,type=type
        ):

        # Bound to unwrapped parent, just try again

        if type(parent) is not _iaw or not hasattr(parent,'REQUEST'):
            return self

        # Create our canonical form
        new_self = _iaw(self,parent)

        if type(self) is _iaw:  # Been wrapped already?
            return new_self     # No special handling


        # Rackmounted? just save/return canonical form
        
        if self._rackMounted:
            self.__dict__[_v_dm_]._registerCanonical(id(self),new_self)
            return new_self


        # Otherwise, find our data manager

        dm = self._DefaultDataManager
        
        try:    _gdmf  = parent.aq_acquire('_getDataManagerFor')
        except: pass
        else:   dm = _gdmf(new_self, dm)
            
        self.__dict__[_v_dm_] = dm
        dm._registerCanonical(id(self),new_self)
        return new_self



    # ObjectManager Hooks to ensure event generation

    def manage_afterAdd(self,item,container):
        self._objectAdding()
        if hasattr(self.aq_base,'objectValues'):
            manage_afterAdd(self,item,container)

    def manage_beforeDelete(self,item,container):
        if hasattr(self.aq_base,'objectValues'):
            manage_beforeDelete(self,item,container)
        self._objectDeleting()


    # Copy Support

    def _getCopy(self, container):
        self = getattr(self,'aq_base',self).__of__(container)
        ob = CopySource._getCopy.im_func(self,container)
        ob.__of__(container)    # force DM hookup
        return ob

    def manage_afterClone(self,item,_v_dm_=_v_dm_):
        self = getattr(self,'aq_base',self)
        self = self.__dict__[_v_dm_]._canonicalForm(id(self),self)
        if hasattr(self.aq_base,'objectValues'):
            manage_afterClone(self,item)















    # Propertysheets
    
    propertysheets = VirtualSheets()

    def _addPropertySheet(self,name='',xmlns='',_v_dm_=_v_dm_):
        """
        Return new propertysheet matching name or xmlns, if possible.
        Otherwise return None.
        """
        d=self.__dict__
        if d.has_key('_v_currentSheets'):
            del d['_v_currentSheets']
        return d[_v_dm_]._addPropertySheetFor(self,name,xmlns)

    def _delPropertySheet(self,name='',xmlns='',_v_dm_=_v_dm_):
        """
        Delete propertysheet matching name or xmlns, if present.  Always
        returns None (unless an unexpected error occurs, of course)
        """
        d=self.__dict__
        if d.has_key('_v_currentSheets'):
            del d['_v_currentSheets']
        d[_v_dm_]._delPropertySheetFor(self,name,xmlns)

    def manage_delete(self,_v_dm_=_v_dm_):
        """Delete this object"""
        if self._rackMounted:
            self.__dict__[_v_dm_]._objectDeleting(self)
            return 'OK'
        raise TypeError, "Not rack-mounted; please use container to delete"

    meta_type = "Data Skin"

    __ac_permissions__ = (
        ('Delete rack-mounted objects', ('manage_delete',)),
    )





    # DEPRECATED STUFF 

    def PropertySheets(self):
        """
        Returns a 'propertysheets'-type object
        
        DEPRECATED - use 'propertysheets' attribute instead.
        """
        return self.propertysheets


    def PropertySheet(self, name='', xmlns=''):
        """
        Return a specific property sheet (DEPRECATED)
        
        Doesn't work w/ZClass sheets!  Use propertysheets.get() instead.
        """
        return self.__dict__[_v_dm_]._PropertySheetFor(self, name, xmlns)

    _v_status_ = ''

default__class_init__(DataSkin)


def initialize(context):
    context.registerBaseClass(DataSkin)

