from PlugIns import PlugIn, PlugInGroup, defaultConstructors
from Transactions import Transactional
from Globals import HTMLFile, default__class_init__
from OFS.PropertyManager import PropertyManager

class Provider(PlugIn, Transactional, PropertyManager):
    """Simple properties-based interface for attribute and sheet providers"""

    manage_options = (
        {'label':'Properties', 'action':'manage_propertiesForm'},
    )

    def manage_editProperties(self, REQUEST):
        """Edit object properties via the web."""
        r = PropertyManager.manage_editProperties(self,REQUEST)
        self._propertiesChanged()
        self.aq_inner.aq_parent.manage_refreshPlugIns()
        return r

    def manage_changeProperties(self, REQUEST=None, **kw):
        """Change existing object properties."""
        r = apply(PropertyManager.manage_changeProperties,(self,REQUEST),kw)
        self._propertiesChanged()
        self.aq_inner.aq_parent.manage_refreshPlugIns()
        return r

    def manage_addProperty(self, id, value, type, REQUEST=None):
        """Add a new property via the web. Sets a new property with
        the given id, type, and value."""
        r = PropertyManager.manage_addProperty(self,id,value,type,REQUEST)
        self._propertiesChanged()
        self.aq_inner.aq_parent.manage_refreshPlugIns()
        return r

    def manage_delProperties(self, ids=None, REQUEST=None):
        """Delete one or more properties specified by 'ids'."""
        r = PropertyManager.manage_delProperties(self,ids,REQUEST)
        self._propertiesChanged()
        self.aq_inner.aq_parent.manage_refreshPlugIns()
        return r

    def _objectAdding(self,client): pass
    def _objectChanging(self,client): pass
    def _objectDeleting(self,client): pass

    def _propertiesChanged(self): pass

    def propertyLabel(self,id):
        return self.propertyLabels.get(id,id)

    propertyLabels={}

default__class_init__(Provider)





























class ProviderGroup(PlugInGroup):
    """
    PlugInGroup that manages the more complex registration needs 
    of Attribute/Sheet/etc. Providers
    """

    def _PlugInsChanged(self, container, plugins=()):

        """Update the registry of names"""

        r = container.__provider_registry__ = {}

        for plugin in plugins:

            nametable = plugin.namesForRegistration(container)

            for namekind, names in nametable.items():
            
                d = r.get(namekind,{})
                
                for n in names:
                    if d.has_key(n):
                        d[n].append(plugin)
                    else:
                        d[n]=[plugin]
                        
                r[namekind] = d

    manage_below = HTMLFile('www/showDiagnostics', globals())

    def manage_registry(self,name):
        return self.aq_parent._getRegistry(name)









    def manage_attrMap(self):
    
        g=self.manage_registry('getattr')
        s=self.manage_registry('setattr')
        d=self.manage_registry('delattr')
        
        m={}; m.update(g); m.update(s); m.update(d)

        for k in m.keys():
            v = m[k] = {'rd':[],'wr':[], 'del':[]}
            if g.has_key(k): v['rd']=g[k]
            if s.has_key(k): v['wr']=s[k]
            if d.has_key(k): v['del']=s[k]

        return m.items()
        
    __ac_permissions__ = (
        ('View management screens',
            ('manage_registry','manage_attrMap','manage_below')
        ),
    )

default__class_init__(ProviderGroup)


















class LinkToParentProviders(Provider):

    """Provide attributes, etc. from an acquired provider"""

    meta_type='Link to parent Data Plug-ins'
    __plugin_kind__ = 'Data Plug-in'

    def namesForRegistration(self,container):

        c=container.aq_parent
        r=getattr(c,'_getRegistry',None)

        if r is None:
            raise TypeError, (
                "No data plug-ins available to link to in /%s" % c.absolute_url(1)
            )

        return {
            'provides': r('provides').keys(),
            
            'getattr': self._handledNames(self.Attributes,r('getattr').keys()),
            'setattr': self._handledNames(self.Attributes,r('setattr').keys()),
            'delattr': self._handledNames(self.Attributes,r('delattr').keys()),

            'sheet_names':
                self._handledNames(
                    self.Sheet_Names, r('sheet_names').keys()
                ),
                
            'sheet_xmlns':
                self._handledNames(
                    self.Sheet_Namespaces, r('sheet_xmlns').keys()
                )
        }







    Attributes = ('*',)
    Sheet_Names = ('*',)
    Sheet_Namespaces = ('*',)

    _properties=(
        {'id':'title', 'type': 'string', 'mode': 'w'},
        {'id':'Attributes', 'type': 'tokens', 'mode': 'w'},
        {'id':'Sheet_Names', 'type': 'tokens', 'mode': 'w'},
        {'id':'Sheet_Namespaces', 'type': 'lines', 'mode': 'w'},
    )

    def _handledNames(self, filt, names):
        if '*' in filt: return names
        if '*' in names:  return filt
        return filter(lambda x,f=filt: x in f, names)


    def _objectAdding(self,client):
        return self.aq_parent.aq_parent._objectAdding(client)

    def _objectDeleting(self,client):
        return self.aq_parent.aq_parent._objectDeleting(client)

    def _objectChanging(self,client):
        return self.aq_parent.aq_parent._objectChanging(client)

    def _checkpoint(self):  return self.aq_parent.aq_parent._checkpoint()
    def _vote(self):        return self.aq_parent.aq_parent._vote()
    def _finalize(self):    return self.aq_parent.aq_parent._finalize()
    def _revert(self):      return self.aq_parent.aq_parent._revert()
    def _rollback(self):    return self.aq_parent.aq_parent._rollback()
    def _cleanup(self):     return self.aq_parent.aq_parent._cleanup()









    def _AttributeFor(self,client,name,default=None):
        return self.aq_parent.aq_parent._AttributeFor(client,name,default)

    def _SetAttributeFor(self,client,name,value):
        return self.aq_parent.aq_parent._SetAttributeFor(client,name,value)

    def _DelAttributeFor(self,client,name):
        return self.aq_parent.aq_parent._DelAttributeFor(client,name)

    def _PropertySheetFor(self,client,name='',xmlns=''):
        return self.aq_parent.aq_parent._PropertySheetFor(client,name,xmlns)

    def _addPropertySheetFor(self,client,name='',xmlns=''):
        return self.aq_parent.aq_parent._addPropertySheetFor(client,name,xmlns)

    def _delPropertySheetFor(self,client,name='',xmlns=''):
        return self.aq_parent.aq_parent._delPropertySheetFor(client,name,xmlns)

    def _PropertySheetsFor(self, client):
        # Note: this could include duplicate sheets if persistent sheets
        # are defined by more than one provider!
        return self.aq_parent.aq_parent._PropertySheetsFor(client)


default__class_init__(LinkToParentProviders)
















def initialize(context):

    context.registerPlugInClass(
        LinkToParentProviders,
        permission = 'Add Links to Parent Data Plug-ins',
        constructors = defaultConstructors(LinkToParentProviders,globals()),
        icon = 'www/acqattrprov.gif'
    )












































